/*
 * Zaptel sysfs support (Internal Details) for Zapata Telephony interface
 *
 * Written by Maxime MENEAULT <maxmeneault@free.fr>
 *
 * Some functions copied from the linux kernel.
 *
 * Copyleft (>) 2007 Maxime MENEAULT <maxmeneault@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
/*
* PRIVATE HEADER for zaptel-sysfs.c 
*
* You can see it as an extension of zaptel-sysfs.c rather than a standard header file.
*
* That's why this header is not protected by ifndef (not meant to be included twice).
*
* This contains all the nasty things about linux kernel versions incompatibilities
* and all the handy macros to handle the linux device model in an easy manner for
* zaptel-sysfs
*
* Some functions (at least one) have been litteraly-copied from newer kernels
* so as to give some functionnality to older ones.
*
* You may add your nasty defines here so as to keep zaptel-sysfs.c clean.
*
* In particular having LINUX_VERSION_CODE checking in zaptel-sysfs.c
* is not allowed!
*/

/* some will be already included but it shouldn't hurt */
#include <linux/kernel.h> /* vsnprintf ... */
#include <linux/device.h> /* linux device model */
#include <linux/sysfs.h> /* don't ask */
#include <linux/stat.h> /* S_IRUGO ... */
#include <linux/kdev_t.h> /* MAJOR ... */
#include <linux/err.h> /* PTR_ERR ... */
#include <linux/kobject.h> /* kobj_uevent_env ... */

/* ===  Various macros that are usefull === */

/* @brief is this span number valid? */
#define ZT_INVALID_SPANNO(_no)\
	((_no < 1) || (_no >= ZT_MAX_SPANS))\

/* @brief is this chan number valid? */
#define ZT_INVALID_CHANNO(_no)\
	((_no < 1) || (_no >= ZT_MAX_CHANNELS))\


/* @brief paste a macro which would result into an identifier
* chunk with another identifier chunk (may be a macro too)
*/
#define PASTE_ID(_macro, _id)	PASTE_ID_P(_macro, _id)
#define PASTE_ID_P(_macro, _id) _macro##_id

/* 
* @brief macro to get an error integer from a pointer 
* (NULL case beeing seen as -ENOMEM)
*/
#define ZT_PTR_ERR(pvoid)\
	((pvoid == NULL)? -ENOMEM : (IS_ERR(pvoid)? PTR_ERR(pvoid) : 0))
	
	
/* === Linux Versionning === */

/* == handling of class_device/device == */

/*
* It was recommended to switch from class_device to device
* much before 2.6.24 but AFAIK some patches dealing with this migration
* were released much after the migration began.
* 2.6.24 should be safe now.
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
#	define USE_DEVICE
#	define zt_device	device
#else
#	define zt_device	class_device
#endif

/* == handling of uevent/hotplug == */

/* fix uevent API changes announced during 2.6.23-rc8 release
* this change it is part of 2.6.24 release
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)


/* @brief declare an uevent callback function
*
*
* @param _name name of the callback to be declared
* @param _dev the name of the zt_device parameter
*@returns the correct prototype for an uevent callback
*/
#	define ZT_UEVENT_CALLBACK(_name, _dev)\
		int _name(struct zt_device *_dev, struct kobj_uevent_env *env)

/* @brief some declarations to be placed at the beginning of every
*  uevent callback definition
*/		
#	define ZT_UEVENT_VAR_DECLS

/*
* @brief add an uevent variable in an uevent environment buffer
*
* the variable will be called ZT_NAME (where NAME is _name)
*
* @param _name name (string)
* @param _fmt  printf-like format string for its value
* @param _val  value (any valid expression matching _fmt)
*
* @warning 
* - to be used within a ZT_UEVENT_CALLBACK function only
* - not to be added into a sub-block without '{' '}'
* -  may exit the function 
*/		
#	define ZT_UEVENT_VAR_ADD(_name, _fmt, _val)\
		if (zt_sysfs_add_uevent_var(env, "ZT_" _name "=" _fmt, (_val)))\
			return -ENOMEM;\

/*
* @brief terminate an uevent environment buffer and exits
*
* @warning: 
* - to be used within a ZT_UEVENT_CALLBACK function only
* - not to be added into a sub-block without '{' '}'
* - exits the function
*/
#	define ZT_UEVENT_VAR_END\
	return 0;
	
#else /* kobject_uevent_env */

#	define ZT_UEVENT_VAR_DECLS\
		int i = 0;\
		int length = 0;
	
#	define ZT_UEVENT_CALLBACK(_name, _dev)\
		int _name(struct zt_device *_dev, char **envp, int num_envp,\
			char *buffer, int buffer_size)


#	define ZT_UEVENT_VAR_ADD(_name, _fmt, _val)\
	{\
		if (zt_sysfs_add_uevent_var(envp, num_envp, &i,\
			buffer, buffer_size, &length,\
			"ZT_" _name "=" _fmt, (_val)))\
			return -ENOMEM;\
	}\


#	define ZT_UEVENT_VAR_END \
	{\
		/* 
		* safety checking
		*/\
		if (i < num_envp) {\
			envp[i] = NULL;\
			return 0;\
		} else\
			return -ENOMEM;\
	}


#endif	/* kobject_uevent_env */


/* fixing other uevent/hotplug issues */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
#	define zt_sysfs_add_uevent_var add_uevent_var

#	ifdef USE_DEVICE
#		define CLASS_UEVENT_FIELD dev_uevent
#	else
#		define CLASS_UEVENT_FIELD uevent
#	endif

#else

/* before 2.6.16 everything was hotplug */
#	define CLASS_UEVENT_FIELD hotplug

#	if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
#		define zt_sysfs_add_uevent_var add_hotplug_env_var
#	else

	/* before kernel 2.6.10 neither add_hotplug_env_var nor add_uevent_var 
	* did exist so here is a copy taken from lib/kobject_uevent.c 
	* (kernel 2.6.19)
	*/
	static int zt_sysfs_add_uevent_var(char **envp, int num_envp, int *cur_index,
		   char *buffer, int buffer_size, int *cur_len,
		   const char *format, ...)
	{
		va_list args;

		/*
	 	* We check against num_envp - 1 to make sure there is at
	 	* least one slot left after we return, since kobject_uevent()
	 	* needs to set the last slot to NULL.
	 	*/
		if (*cur_index >= num_envp - 1)
			return -ENOMEM;

		envp[*cur_index] = buffer + *cur_len;

		va_start(args, format);
		*cur_len += vsnprintf(envp[*cur_index],
			      max(buffer_size - *cur_len, 0),
			      format, args) + 1;
		va_end(args);

		if (*cur_len > buffer_size)
			return -ENOMEM;

		(*cur_index)++;
		return 0;
	}
#	endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10) */
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16) */


/* class didn't have owner member before 2.6.13 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
#	define CLASS_OWNER(...) __VA_ARGS__
#else
#	define CLASS_OWNER(...)
#endif

/* before 2.6.12 devt member of class_device didn't exist */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
#	define CLASS_DEVICE_HAVE_DEVT
#endif

/* before 2.6.15 release member of class_device didn't exist 
* (only needed if we want to override class release function)
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
#	define CLASS_DEVICE_RELEASE(...) __VA_ARGS__
#else
#	define CLASS_DEVICE_RELEASE(...)
#endif


/* == handling of class_device/device (cont.) == */

#ifndef USE_DEVICE

#	define DEVICE_NAME_FIELD	class_id
#	define CLASS_RELEASE_FIELD	release
#	define DEVICE_DATA_FIELD	class_data
#	define DEVICE_DEVICE_FIELD  dev

/* @brief get the show function prototype for zt_device_attribute(s)
*
* @param _name name of the funtion
* @param _dev name of the zt_device parameter
* @param _buf name of the buffer parameter
* @returns the show function prototyp
*/
#	define ATTR_SHOW_FUNC(_name, _dev, _buf)\
	ssize_t _name(struct zt_device *_dev, char *_buf)
		
#else

#	define DEVICE_NAME_FIELD	bus_id
#	define DEVICE_DATA_FIELD	driver_data
#	define DEVICE_DEVICE_FIELD  parent

/* be aware that you shouldn't use class_device anymore under
* the zaptel class for version with USE_DEVICE defined otherwise
* you may get an error if the class does not contain a release function
* (to release class_device(s))
*/
#	define CLASS_RELEASE_FIELD	dev_release

#	define ATTR_SHOW_FUNC(_name, _dev, _buf)\
	ssize_t _name(struct zt_device *_dev, struct zt_device_attribute *attr,\
		char *_buf)
	
#endif /* USE_DEVICE */

#define zt_device_attribute PASTE_ID(zt_device, _attribute)
#define zt_device_create_file PASTE_ID(zt_device, _create_file)
#define zt_device_remove_file PASTE_ID(zt_device, _remove_file)


/* === zt_device handling === */

/*
*@brief a wrapper structure arround a zt_device to remember some 
* interesting things
*/
struct zt_sysfs_device {
	int is_static; /*!< did we allocate it with kmalloc or is it static memory? */
	int type; 	  /*!< type of this class device one of ZT_DEVICE_TYPE_* */
	struct zt_device device; /*!< the device */
	  
#ifndef CLASS_DEVICE_HAVE_DEVT
	dev_t devt;	/*!< for older kernels */
#endif
};

#define to_zt_sysfs_device(_pdev)\
	container_of(_pdev, struct zt_sysfs_device, device)

/* === class_device dev_t member handling === */

/*
* This gives us the compatibility even before simple_class was added.
*/
#ifdef CLASS_DEVICE_HAVE_DEVT

#	define CLASS_DEVICE_DEVT_HAVE(...) __VA_ARGS__

#	define CLASS_DEVICE_DEVT_HAVENT(...)

#	define zt_device_register(zdev)\
		PASTE_ID(zt_device, _register)(&((zdev)->device))

#	define zt_device_unregister(zdev)\
		PASTE_ID(zt_device, _unregister)(&((zdev)->device))
#else

#	define CLASS_DEVICE_DEVT_HAVE(...)

#	define CLASS_DEVICE_DEVT_HAVENT(...) __VA_ARGS__

static ATTR_SHOW_FUNC(zt_show_dev_t, device, buf)
{
	struct zt_sysfs_device *zdev = NULL;
	/* maybe unecessary checking but doesn't hurt */
	if (!device)
		return 0;
	zdev = to_zt_sysfs_device(device);
	return print_dev_t(buf, zdev->devt);
}
/* Apparently "udev" needs a file called "dev" under each class_device representing
* a device that should go into /dev
*/
static struct zt_device_attribute zt_device_attr_devt =
{
	.attr = {.name = "dev", .owner = THIS_MODULE, .mode = S_IRUGO},
	.show = zt_show_dev_t,
};

static int zt_device_register(struct zt_sysfs_device *zdev)
{
	int retval = class_device_register(&zdev->device);
    
	if (retval)
		return retval;
        
	if (MAJOR(zdev->devt)) {
		retval = class_device_create_file(&zdev->device, &zt_device_attr_devt);
		if (retval)
			goto error;
	}
    
	return 0;
error:
	class_device_unregister(&zdev->device);
	return retval;
}

static void zt_device_unregister(struct zt_sysfs_device *zdev)
{
	if (zdev) {
		class_device_remove_file(&zdev->device, &zt_device_attr_devt);
		class_device_unregister(&zdev->device);
	}
}
#endif /* CLASS_DEVICE_HAVE_DEVT */



/*
* @brief function to release the the only class of zaptel 
* (it is a no-op because the class is static memory)
*/
static void zt_sysfs_class_release_nop(struct class *cls)
{
}

/* @brief function to release zt_device(s) under the zaptel class */
static void zt_sysfs_device_release(struct zt_device *device)
{
	struct zt_sysfs_device *zdev = NULL;
    
	if (!device)
		return;
    
	zdev = to_zt_sysfs_device(device);
    
	if (zdev->is_static)
		return;
        
	kfree(zdev);
}



/* === Statically allocated zt_device(s) handling === */

/* @brief To terminate an array of zt_device(s) */
#define ZT_STATIC_DEVICE_NULL \
	{.device = {.class = NULL}}

/* @brief Add a static zt_device to an array 
* 
* This zt_device will be under the main zaptel class
* (for kernels using struct device it is likely that these devices
*  will be located under /sys/devices/virtual because they don't
*  have any real parent)
*/     
#define ZT_DEFINE_STATIC_DEVICE(_name, _dev)\
	{\
		.is_static = 1,\
		.type = ZT_DEVICE_TYPE_DEFAULT,\
		.device = {\
			.DEVICE_NAME_FIELD = #_name,\
			.class = &zt_sysfs_class,\
			CLASS_DEVICE_DEVT_HAVE(.devt = _dev,)\
		},\
		CLASS_DEVICE_DEVT_HAVENT(.devt = _dev,)\
}\



/* === Dynamically allocated zt_device handling === */

/* simple declaration before the real definition (to avoid undefined errors) */
static struct class zt_sysfs_class;

/*
* @brief create a dynamically allocated zaptel zt_device
* 
* @param dev the underlying device if known
* @param devt the dev_t type if known or meaningfull
* @param data depends on type (for the moment either a zt_span or a zt_chan or NULL)
* @param type defines the type of this device (one of ZT_DEVICE_*)
* @param fmt  printf-like format string to give a name to this zt_device
* @param ...  variables argument list that match fmt
*
* @return the pointer to newly created device or an error pointer
*/
static struct zt_sysfs_device *zt_sysfs_device_create(struct device *dev,
	dev_t devt, void *data, int type, const char *fmt, ...)
{
	va_list args;
	struct zt_sysfs_device *zdev = NULL;
	int retval = -ENODEV;

	zdev = kmalloc(sizeof(*zdev), GFP_KERNEL);
    
	if (!zdev) {
		retval = -ENOMEM;
		goto error;
	}
    
	memset(zdev, 0, sizeof(*zdev));
	/* be sure to set is_static field as 0 */
    
	zdev->type = type;
    
	if (MAJOR(devt)) {
		CLASS_DEVICE_DEVT_HAVE(zdev->device.devt = devt);
		CLASS_DEVICE_DEVT_HAVENT(zdev->devt = devt);
	}
    
	zdev->device.DEVICE_DEVICE_FIELD = dev;
	zdev->device.class = &zt_sysfs_class;
    
	zdev->device.DEVICE_DATA_FIELD = data;
    
	va_start(args, fmt);
	vsnprintf(zdev->device.DEVICE_NAME_FIELD, BUS_ID_SIZE, fmt, args);
	va_end(args);
    
	retval = zt_device_register(zdev);
	if (retval)
		goto error;

	return zdev;
error:
	kfree(zdev);
	return ERR_PTR(retval);
}


/* === Class device attributes handling === */

/*
* @brief get the name of function show (for an attribute)
*        it avoids name space clashes
*
* @param _name a base name that is unique for _type
* @param _type either chan or span
* 
*/  
#define ZT_ATTR_GET_SHOW(_name, _type)\
	zt_sysfs_##_type##_show_##_name
      
/*
* @brief define a show function (to use in a class_device attribute)
* for either a span or channel.
* 
* This only handles attributes to print a single expression result.
*
* @param _name name of this attribute
* @param _type either chan or span
* @param _typeup _type in upper-case
* @param _vartype the C type of the variable that can hold _expr
* @param _fmt the printf-like format string to print a _vartype
* @param _expr the expression that should be printed
*
* @note the resulting function has got the name ZT_ATTR_GET_SHOW(_name, _type)
*
* @note the _expr is executed under the chan->lock for a channel and under
*       span->lock for a span.
*
* @note class_device attributes are an overkill (for now -- kernel 2.6.19)
*       we have to define a separate function for each attribute...
*/
#define ZT_ATTR_DEFINE_SHOW_SIMPLE(_name, _type, _typeup, _vartype, _fmt, _expr)\
	static ATTR_SHOW_FUNC(ZT_ATTR_GET_SHOW(_name, _type), device, buf)\
	{\
		struct zt_##_type *_type = (struct zt_##_type *)\
				device->DEVICE_DATA_FIELD;\
		unsigned long flags;\
		_vartype var;\
		\
		if (!_type) {\
			return 0;\
		}\
		\
		spin_lock_irqsave(&_type->lock, flags);\
		\
		if (ZT_INVALID_##_typeup##NO(_type->_type##no)) {\
			spin_unlock_irqrestore(&_type->lock, flags);\
			return 0;\
		}\
		\
		var = (_vartype)(_expr);\
		spin_unlock_irqrestore(&_type->lock, flags);\
		\
		return sprintf(buf, _fmt, var);\
	}
    

   
/*
* @brief define a read-only attribute that uses a show function
*        which name has been defined with ZT_ATTR_GET_SHOW
*/
#define ZT_ATTR_DEFINE_RO(_name, _type)\
	{\
		.attr = {.name = #_name, .mode = S_IRUGO, .owner = THIS_MODULE},\
		.show = ZT_ATTR_GET_SHOW(_name, _type),\
    }

/*
* @brief the NULL attribute (to terminate an array of zt_device attributes)
*/    
#define ZT_ATTR_NULL \
	{.attr = {.name = NULL}}


/*
* @brief define a show function for a member of the zt_chan structure
*
* @param _member the member of the structure to print
* @param _vartype its C type
* @param _fmt a printf-like format that can print _vartype
*
* @see ZT_ATTR_DEFINE_SHOW_SIMPLE
*/
#define ZT_ATTR_CHAN_SHOW_MEMBER(_member, _vartype, _fmt)\
	ZT_ATTR_DEFINE_SHOW_SIMPLE(_member, chan, CHAN, _vartype, _fmt, chan->_member)
    
/*
* @brief define a read-only attribute that use a show function
*		which name has been defined with ZT_ATTR_GET_SHOW(*, chan)
*/
#define ZT_ATTR_CHAN_RO(_name)\
	ZT_ATTR_DEFINE_RO(_name, chan)

/*
* @brief define a show function for a member of the zt_span structure
*
* @param _member the member of the structure to print
* @param _vartype its C type
* @param _fmt a printf-like format that can print _vartype
*
* @see ZT_ATTR_DEFINE_SHOW_SIMPLE
*/
#define ZT_ATTR_SPAN_SHOW_MEMBER(_member, _vartype, _fmt)\
	ZT_ATTR_DEFINE_SHOW_SIMPLE(_member, span, SPAN, _vartype, _fmt, span->_member)
    
/*
* @brief define a read-only attribute that use a show function
*		which name has been defined with ZT_ATTR_GET_SHOW(*, span)
*/
#define ZT_ATTR_SPAN_RO(_name)\
	ZT_ATTR_DEFINE_RO(_name, span)
